﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.BioBoost.Representation {

  public interface ILayerDescriptor : IItem {
    string Name { get; }
    string Unit { get; }
    string Description { get; }
    string Separator { get; }
    string FullName { get; }
  }

  public interface IValueLayer {
    LayerDescriptor LayerDescriptor { get; }
    double this[int idx] { get; set; }
  }

  public interface IAggregator {
    IEnumerable<IValueLayer> Aggregate(IEnumerable<IValueLayer> layers, ItemArray<ItemList<IntValue>> sources);
  }

  public class SumAggregator : IAggregator {
    public ILayerDescriptor TargetLayerDescriptor { get; private set; }
    public SumAggregator(ILayerDescriptor descriptor) { TargetLayerDescriptor = descriptor; }
    public IEnumerable<IValueLayer> Aggregate(IEnumerable<IValueLayer> layers, ItemArray<ItemList<IntValue>> sources) {
      foreach (var layer in layers) {
        if (layer.LayerDescriptor == TargetLayerDescriptor) {
          //yield return 
        }
      }
      yield break;
    }
  }

  [StorableClass(StorableClassType.AllFieldsAndAllProperties)]
  public class LayerDescriptor : Item, ILayerDescriptor {

    public string Name  { get; private set; }
    public string Unit { get; private set; }
    public string Description { get; private set; }
    public string Separator { get; private set; }
    public string FullName { get; private set; }

    protected LayerDescriptor(bool isDeserializing) : base(isDeserializing) { }
    protected LayerDescriptor(LayerDescriptor orig, Cloner cloner) : base(orig, cloner) { }

    public LayerDescriptor(string name, string unit, string separator = " ") {
      Name = name;
      Unit = unit;
      Separator = separator;
      FullName = separator + name;
    }

    public string NameWithPrefix(string prefix) { return string.Format("{0}{1}", prefix, FullName); }
    public bool IsSuffixOf(string s) { return s.EndsWith(FullName); }
    public string RemoveSuffixFrom(string s) {
      if (!s.EndsWith(FullName)) return s;
      return s.Substring(0, s.Length - FullName.Length);
    }

    public override IDeepCloneable Clone(Cloner cloner) { return new LayerDescriptor(this, cloner); }

    public static LayerDescriptor AmountsAtSource = new LayerDescriptor("amounts at source", "t/a");
    public static LayerDescriptor AmountsAtTarget = new LayerDescriptor("amounts at target", "t/a");
    public static LayerDescriptor AmountsTransportedFromSource = new LayerDescriptor("amounts transported from source", "t/a");
    public static LayerDescriptor PotentialsFromProblemData = new LayerDescriptor("Potentials", "t/a", "");
    public static LayerDescriptor Utilizations = new LayerDescriptor("Utilizations", "");
    public static LayerDescriptor UtilizationsEffective = new LayerDescriptor("utilizations (effective)", "");
    public static LayerDescriptor TransportTargets = new LayerDescriptor("Transport Targets", "region");
    public static LayerDescriptor ConverterCapacities = new LayerDescriptor("converter capacities", "t/a");
    public static LayerDescriptor StorageCapacities = new LayerDescriptor("storage capacities", "t/a");
    public static LayerDescriptor AcquisitionCosts = new LayerDescriptor("acquisition costs", "€/a");
    public static LayerDescriptor DischargeCosts = new LayerDescriptor("discharge costs", "€/a");
    public static LayerDescriptor TransportCosts = new LayerDescriptor("transport costs", "€/a");
    public static LayerDescriptor TransportCostsPerT = new LayerDescriptor("transport costs (per t)", "€/t");
    public static LayerDescriptor TransportCostsPerTkm = new LayerDescriptor("transport costs (per tkm)", "€/tkm");
    public static LayerDescriptor HandlingCosts = new LayerDescriptor("handling costs", "€/a");
    public static LayerDescriptor RelativeHandlingCosts = new LayerDescriptor("handling costs (relative)", "€/t");
    public static LayerDescriptor TransportModes = new LayerDescriptor("transport modes", "mode");
    public static LayerDescriptor ConversionCosts = new LayerDescriptor("conversion costs", "€/a");
    public static LayerDescriptor TotalAmortizedConversionCosts = new LayerDescriptor("total amortized conversion costs (incl. OPEX, CAPEX)", "€/a");
    public static LayerDescriptor RelativeTotalAmortizedConversionCosts = new LayerDescriptor("total amortized conversion costs (incl. OPEX, CAPEX, relative)", "€/t");
    public static LayerDescriptor ConversionPlantConstructionCost = new LayerDescriptor("conversion plant construction costs", "€/a");
    public static LayerDescriptor ConversionPlantOperationCost = new LayerDescriptor("conversion plant operation costs", "€/a");
    public static LayerDescriptor RelativeConversionCosts = new LayerDescriptor("conversion costs (relative)", "€/t");
    public static LayerDescriptor RelativeConversionPlantConstructionCosts = new LayerDescriptor("conversion plant construction costs (relative)", "€/t");
    public static LayerDescriptor RelativeConversionPlantOperationCosts = new LayerDescriptor("conversion plant operation costs (relative)", "€/t");
    public static LayerDescriptor StorageCost = new LayerDescriptor("storage costs", "€/a");
    public static LayerDescriptor RelativeStorageCost = new LayerDescriptor("storage costs (relative)", "€/t");
    public static LayerDescriptor TransportDistance = new LayerDescriptor("transport distances", "km");
    public static LayerDescriptor ExceedingTransportDistancePenalty = new LayerDescriptor("transport penalties", "");
    public static LayerDescriptor Tkm = new LayerDescriptor("tkm", "tkm");
    public static LayerDescriptor MaxConversionCapacity = new LayerDescriptor("converter capacities (max)", "t/a");
    public static LayerDescriptor ScalingPenalty = new LayerDescriptor("converter scaling penalties", "");
    public static LayerDescriptor TotalCostsAtTarget = new LayerDescriptor("total costs at target", "€/a");
    public static LayerDescriptor TotalCostsAtSource = new LayerDescriptor("total costs at source", "€/a");
    public static LayerDescriptor LocalAddedValue = new LayerDescriptor("local added value", "€/a", "");
    public static LayerDescriptor RelativeCostAtSource = new LayerDescriptor("total costs at source (relative)", "€/t");
    public static LayerDescriptor RelativeCostAtTarget = new LayerDescriptor("total costs at target (relative)", "€/t");
    public static LayerDescriptor AmountsAtTargetConverted = new LayerDescriptor("amounts converted at target", "t/a");
    public static LayerDescriptor TotalLogisticCosts = new LayerDescriptor("logistics costs", "€/a");
    public static LayerDescriptor LogisticCostsPerTon = new LayerDescriptor("logistics costs (per t)", "€/t");
    public static LayerDescriptor LogisticCostsPerTonKm = new LayerDescriptor("logistics costs (per tkm)", "€/tkm");
    public static LayerDescriptor TotalDirectConversionCost = new LayerDescriptor("conversion costs (incl. supplies)", "€/a");
    public static LayerDescriptor RelativeTotalDirectConversionCost = new LayerDescriptor("conversion costs (incl. supplies, relative)", "€/t");
    public static LayerDescriptor TotalPenalty = new LayerDescriptor("total penalties", "", "");
    public static LayerDescriptor Sources = new LayerDescriptor("conversion plant sources", "");
    public static LayerDescriptor ScalingFactors = new LayerDescriptor("conversion plant scaling factors", "");
    public static LayerDescriptor ExceedingCapacityPenalties = new LayerDescriptor("exceeding capacity penalties", "");

    //public static LayerDescriptor AverageUtilizations = new LayerDescriptor("average utilizations", "t/a");
    //public static LayerDescriptor AverageAmountsFromSource = new LayerDescriptor("average amounts from source", "t/a");
    //public static LayerDescriptor AggregateAmountsFromSource = new LayerDescriptor("aggregate amounts from source", "t/a");
    //public static LayerDescriptor AverageCostsFromSource = new LayerDescriptor("average costs from source", "t/a");
    //public static LayerDescriptor AggregateCostsFromSource = new LayerDescriptor("aggregate costs from source", "t/a");

    public static LayerDescriptor FeedstockCosts = new LayerDescriptor("feedstock costs", "€/a");
    public static LayerDescriptor FeedstockCostsRelative = new LayerDescriptor("feedstock costs (relative)", "€/t");
    //public static LayerDescriptor AggregatedAverageFeedstockCosts = new LayerDescriptor("aggregated average feedstock costs", "€/a");
    //public static LayerDescriptor AggregatedFeedstockTransportCosts = new LayerDescriptor("aggregated feedstock transport costs", "€/a");
    //public static LayerDescriptor RelativeAggregatedFeedstockTransportCosts = new LayerDescriptor("aggregated feedstock transport costs (relative)", "€/t");
    //public static LayerDescriptor AggregatedFeedstockHandlingCosts = new LayerDescriptor("aggregated feedstock handling costs", "€/a");
    //public static LayerDescriptor RelativeAggregatedFeedstockHandlingCosts = new LayerDescriptor("aggregated feedstock handling costs (relative)", "€/t");
    //public static LayerDescriptor AggregatedFeedstockStorageCosts = new LayerDescriptor("aggregated feedstock storage costs", "€/a");
    //public static LayerDescriptor RelativeAggregatedFeedstockStorageCosts = new LayerDescriptor("aggregated feedstock storage costs (relative)", "€/t");
    public static LayerDescriptor DecentralLogisticCosts = new LayerDescriptor("logistic costs (decentral aggregated)", "€/a");
    public static LayerDescriptor DecentralLogisticCostsRelative = new LayerDescriptor("logistic costs (decentral aggregated, relative)", "€/t");
    //public static LayerDescriptor AggregatedDirectConversionCosts = new LayerDescriptor("aggregated direct conversion costs", "€/a");
    //public static LayerDescriptor RelativeAggregatedDirectConversionCosts = new LayerDescriptor("aggregated direct conversion costs (relative)", "€/t");
    //public static LayerDescriptor AggregatedConversionOperationCosts = new LayerDescriptor("aggregated conversion operation costs", "€/a");
    //public static LayerDescriptor RelativeAggregatedConversionOperationCosts = new LayerDescriptor("aggregated conversion operation costs (relative)", "€/t");
    //public static LayerDescriptor AggregatedConversionConstructionCosts = new LayerDescriptor("aggregated conversion construction costs", "€/a");
    //public static LayerDescriptor RelativeAggregatedConversionConstructionCosts = new LayerDescriptor("aggregated conversion construction costs (relative)", "€/t");
    public static LayerDescriptor DecentralConversionCosts = new LayerDescriptor("conversion costs (decentral aggregated)", "€/a");
    public static LayerDescriptor DecentralConversionCostsRelative = new LayerDescriptor("conversion costs (decentral aggregated, relative)", "€/t");
    public static LayerDescriptor CentralConversionCostsRelative = new LayerDescriptor("conversion costs (central aggregated, relative)", "€/t");
    // ...
    public static LayerDescriptor CentralLogisticCosts = new LayerDescriptor("logistic costs (central aggregated)", "€/a");
    public static LayerDescriptor CentralLogisticCostsRelative = new LayerDescriptor("logistic costs (central aggregated, relative)", "€/t");

    public static LayerDescriptor RoiFollowed = new LayerDescriptor("roi (followed)", "€/€");
    public static LayerDescriptor RoiAggregated = new LayerDescriptor("roi (aggregated)", "€/€");



    private static readonly List<LayerDescriptor> LAYERS = new List<LayerDescriptor>();

    private static void DiscoverLayers() {
        if (LAYERS.Count == 0) {
          LAYERS.AddRange(
            typeof (LayerDescriptor)
              .GetFields(BindingFlags.Static | BindingFlags.Public)
              .Select(f => f.GetValue(null))
              .Where(v => v is LayerDescriptor)
              .Cast<LayerDescriptor>()
              .OrderByDescending(l => l.Name.Length));
          foreach (var l1 in LAYERS) {
            foreach (var l2 in LAYERS) {
              if (l1 != l2 && l1.IsSuffixOf(l2.FullName))
                throw new InvalidOperationException(string.Format("Layer overlap detected \"{0}\" is suffix of \"{1}\"", l1.FullName, l2.FullName));
            }
          }
        }
    }

    public static IEnumerable<LayerDescriptor> Layers {
      get {  DiscoverLayers(); return LAYERS; }
    }

    public static LayerDescriptor GetLayer(string fullName) {
      DiscoverLayers();
      foreach (var layer in LAYERS) {
        if (layer.IsSuffixOf(fullName)) return layer;
      }
      return null;
    }
  }
}
